class Canonical:

  def is_canonical(self):
    """Checks if a containment line is in the canonical form.

    As all containments (GFA1 C lines) can be specified using one of the
    two sequences in the positive orientation, only one of the possible
    variants is considered canonical by Gfapy; this allow to check if
    two C lines are equivalent to each other. In particular, the C line
    is considered canonical, if the from_orient is +.

    Note:
      An example: a containment of B (length:8) in A (length:100) at
      position 9 of A, with a cigar 1M1I2M3D4M (i.e. rpos = 19).

      ::
          A+ B+ 1M1I2M3D4M 9 == A- B- 4M3D2M1I1M 80
          A+ B- 1M1I2M3D4M 9 == A- B+ 4M3D2M1I1M 80
          A- B+ 1M1I2M3D4M 9 == A+ B- 4M3D2M1I1M 80
          A- B- 1M1I2M3D4M 9 == A+ B+ 4M3D2M1I1M 80

      Pos in the complement is equal to the length of A minus the right pos
      of B before reversing.
      We require here that A != B as A == B makes no sense for containments.
      Thus it is always possible to express the containment using a positive
      from orientation.

    Returns:
      bool
    """
    return self.from_orient() == "+"



import gfapy
import re

def decode(string):
  return gfapy.Alignment(string, valid = False, version = "gfa1")

def unsafe_decode(string):
  return gfapy.Alignment(string, valid = True, version = "gfa1")

def validate_encoded(string):
  if not re.match(r"^(\*|([0-9]+[MIDNSHPX=])+)$", string):
    raise gfapy.FormatError(
      "{} is not a valid GFA1 alignment\n".format(repr(string)) +
      "(it is not * and is not a CIGAR string (([0-9]+[MIDNSHPX=])+)")

def validate_decoded(obj):
  if isinstance(obj, gfapy.CIGAR):
    obj.validate()
  elif isinstance(obj, gfapy.Placeholder):
    pass
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name) +
      "(accepted classes: gfapy.CIGAR, gfapy.Placeholder)")

def unsafe_encode(obj):
  return str(obj)

def encode(obj):
  if isinstance(obj, str):
    validate_encoded(obj)
    return obj
  elif isinstance(obj, gfapy.CIGAR):
    obj.validate()
    return str(obj)
  elif isinstance(obj, gfapy.Placeholder):
    return str(obj)
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__) +
      "(accepted classes: str, gfapy.CIGAR, gfapy.Placeholder)")



import gfapy

class Construction:
  def _initialize_positional_fields(self, strings):
    self._init_field_value("content", "comment", strings[1], errmsginfo = strings)
    sp = strings[2] if len(strings) > 2 else " "
    self._init_field_value("spacer", "comment", sp, errmsginfo = strings)

  def _initialize_tags(self, strings):
    if len(strings) > 3:
      raise gfapy.ValueError("Comment lines do not support tags")



import gfapy
import unittest

class TestAPIPositions(unittest.TestCase):

  def test_positions(self):
    # from string and integer
    pos1 = gfapy.LastPos(12); pos2 = gfapy.LastPos("12$")
    self.assertEqual(pos1, pos2)
    assert(isinstance(pos1, gfapy.LastPos))
    assert(isinstance(pos2, gfapy.LastPos))
    # value
    self.assertEqual(12, gfapy.posvalue(pos1))
    self.assertEqual(12, gfapy.posvalue(pos2))
    self.assertEqual(12, gfapy.posvalue(12))
    # to_pos on string without dollar
    self.assertEqual(12, gfapy.LastPos("12"))
    assert(isinstance(gfapy.LastPos("12"), int))
    # to pos: wrong format
    with self.assertRaises(gfapy.FormatError): gfapy.LastPos("12=")
    # 0$ is allowed, although unclear if useful
    assert(gfapy.islastpos(gfapy.LastPos("0$")))
    # comparison with integer and string
    self.assertEqual(gfapy.LastPos(10), 10)
    self.assertEqual(10, gfapy.LastPos(10))
    # to_s
    self.assertEqual("12$", str(pos1))
    # to_i
    self.assertEqual(12, int(pos1))

  def test_positions_negative(self):
    # negative values
    with self.assertRaises (gfapy.ValueError): gfapy.LastPos("-1")
    with self.assertRaises (gfapy.ValueError): gfapy.LastPos("-1$")
    # negative values, valid: True
    self.assertEqual(-1, gfapy.LastPos("-1",valid=True))
    assert(isinstance(gfapy.LastPos("-1",valid=True), int))
    self.assertEqual(gfapy.LastPos(-1, valid=True), gfapy.LastPos("-1$",valid=True))
    self.assertEqual(gfapy.LastPos(-1, valid=True), gfapy.LastPos(-1,valid=True))
    # validate
    with self.assertRaises (gfapy.ValueError): gfapy.LastPos("-1$",valid=True).validate()
    with self.assertRaises (gfapy.ValueError): gfapy.LastPos(-1,valid=True).validate()

  def test_positions_first_last(self):
    assert(not gfapy.islastpos(gfapy.LastPos("0")))
    assert(not gfapy.islastpos(gfapy.LastPos("12")))
    assert(gfapy.islastpos(gfapy.LastPos("12$")))
    assert(gfapy.isfirstpos(gfapy.LastPos("0")))
    assert(not gfapy.isfirstpos(gfapy.LastPos("12")))
    assert(not gfapy.isfirstpos(gfapy.LastPos("12$")))

  def test_positions_subtract(self):
    a = gfapy.LastPos("13$")
    a1 = a - 0
    a2 = a - 1
    self.assertEqual(13, a1)
    self.assertEqual(12, a2)
    assert(gfapy.islastpos(a1))
    assert(not gfapy.islastpos(a2))




import gfapy

class GFA1ToGFA2:

  def _to_gfa2_a(self):
    """
    Returns
    -------
    list of str
    	A list of GFA2 field strings.
    """
    try:
       length = self.try_get_length()
    except gfapy.NotFoundError:
      raise gfapy.RuntimeError(
          "Conversion of GFA1 segment line to GFA2 failed\n"+
          "GFA2 requires to specify a length\n"+
          "No length information available in the GFA1 segment:\n"+
          "Segment line: {}".format(str(self)))
    a = ["S", self.field_to_s("name", tag = False), str(self.try_get_length()),
              self.field_to_s("sequence", tag = False)]
    for fn in self.tagnames:
      if fn != "LN":
        a.append(self.field_to_s(fn, tag = True))
    return a



import gfapy
import unittest

class TestInternalsFieldParser(unittest.TestCase):

  def test_parse_gfa_tag(self):
    o = "AA:i:1"
    self.assertEqual(["AA","i","1"], gfapy.Field._parse_gfa_tag(o))
    with self.assertRaises(gfapy.FormatError):
      gfapy.Field._parse_gfa_tag("1A:A:A")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Field._parse_gfa_tag("_A:A:A")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Field._parse_gfa_tag("A:A:A")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Field._parse_gfa_tag("AAA:A:A")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Field._parse_gfa_tag("AA:C:1")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Field._parse_gfa_tag("AA:AA:1")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Field._parse_gfa_tag("AA:a:1")

  def test_parse_gfa_field_A(self):
    self.assertEqual("1", gfapy.Field._parse_gfa_field("1", "A"))

  def test_parse_gfa_field_i(self):
    self.assertEqual(12, gfapy.Field._parse_gfa_field("12", "i"))

  def test_parse_gfa_field_f(self):
    self.assertEqual(1.2, gfapy.Field._parse_gfa_field("1.2", "f"))

  def test_parse_gfa_field_Z(self):
    self.assertEqual("1.2", gfapy.Field._parse_gfa_field("1.2", "Z"))

  def test_parse_gfa_field_H(self):
    self.assertEqual(gfapy.ByteArray([26]),
        gfapy.Field._parse_gfa_field("1A", "H"))

  def test_parse_gfa_field_B(self):
    self.assertEqual([12,12,12],
        gfapy.Field._parse_gfa_field("c,12,12,12", "B"))
    self.assertEqual([1.2,1.2,1.2],
        gfapy.Field._parse_gfa_field("f,1.2,1.2,1.2", "B"))

  def test_parse_gfa_field_J(self):
    self.assertEqual({"1" : 2},
        gfapy.Field._parse_gfa_field("{\"1\":2}", "J"))



import unittest
import gfapy

class TestAlignment(unittest.TestCase):

  def test_string_to_cigar(self):
    self.assertEqual(gfapy.CIGAR([
      gfapy.CIGAR.Operation(12, "M"),
      gfapy.CIGAR.Operation(1,  "D"),
      gfapy.CIGAR.Operation(2,  "I"),
      ]), gfapy.Alignment("12M1D2I"))

  def test_string_to_placeholder(self):
    self.assertIsInstance(gfapy.Alignment("*"), gfapy.Placeholder)

  def test_string_to_trace(self):
    self.assertEqual(gfapy.Trace([12,14,15]),
                     gfapy.Alignment("12,14,15"))

  def test_string_invalid(self):
    self.assertRaises(gfapy.FormatError,
                      gfapy.Alignment, "12x1,D2I")

  def test_list_to_cigar(self):
    self.assertEqual(gfapy.CIGAR([
      gfapy.CIGAR.Operation(12, "M"),
      gfapy.CIGAR.Operation(1,  "D"),
      gfapy.CIGAR.Operation(2, "I")]),
      gfapy.Alignment(
      [gfapy.CIGAR.Operation(12, "M"),
       gfapy.CIGAR.Operation(1,  "D"),
       gfapy.CIGAR.Operation(2,  "I")]))

  def test_list_to_trace(self):
    self.assertEqual(gfapy.Trace([12,14,15]),
                     gfapy.Alignment([12,14,15]))

  def test_list_invalid(self):
    self.assertRaises(gfapy.FormatError,
                      gfapy.Alignment,["12x1", "2I"])



from collections import OrderedDict
from functools import partial
import re
import gfapy
from .dynamic_fields import DynamicField

try:
  from functools import partialmethod
except ImportError:
  #for compatibility with old python versions
  def partialmethod(method, **kwargs):
    return lambda self: method(self, **kwargs)

class Construction:

  DELAYED_PARSING_DATATYPES = [
                                "alignment_gfa1",
                                "alignment_gfa2",
                                "alignment_list_gfa1",
                                "oriented_segments",
                                "H",
                                "J",
                                "B",
                              ]
  """
  List of datatypes which are parsed only on access.

  All other are parsed when read.
  """

  RECORD_TYPE_VERSIONS = {
      "specific" :
        {"gfa1" : ["C", "L", "P"],
         "gfa2" : ["E", "G", "F", "O", "U", "\n"]},
      "generic" : ["H", "#"],
      "different" : ["S"]
    }
  """
  Dependency of record type from version.

  * specific: only for a specific version
  * generic:  same syntax for all versions
  * different: different syntax in different versions
  """

  def __new__(cls, data, vlevel = 1, virtual = False, dialect = "standard",
      version = None):
    if isinstance(data, str):
      data = data.split("\t")
    if isinstance(data, list) and cls.RECORD_TYPE == None:
      cls = gfapy.Line._subclass(data, version = version)
    return object.__new__(cls)

  def __init__(self, data, vlevel = 1, virtual = False,
               version = None, dialect = "standard"):
    self._dialect = dialect.lower()
    self.vlevel = vlevel
    self._virtual = virtual
    self._datatype = {}
    self._data = {}
    self._gfa = None
    self._version = version
    self._refs = {}
    if self.__class__ == gfapy.Line:
      raise gfapy.AssertionError("Line subclass unknown")
    if isinstance(data, dict):
      # API private initialization using dict
      self._data.update(data)
    else:
      # public initialization using list (or tab-separated string)
      if self.__class__ == gfapy.line.Comment:
        data = gfapy.Line._init_comment_data(data)
      elif isinstance(data, str):
        data = data.split(gfapy.Line.SEPARATOR)
      if self.version is None:
        self._compute_version(data[0])
      else:
        self._validate_version()
      self._initialize_positional_fields(data)
      self._initialize_tags(data)
      if self.vlevel >= 1:
        self._validate_record_type_specific_info()
      if self.version is None:
        raise gfapy.RuntimeError("version could not be determined, "+
            "record_type={}".format(self.record_type))

  @staticmethod
  def _init_comment_data(data):
    if isinstance(data, list) and (data[0] != "#"):
      # unproperly splitten, rejoin
      data = "\t".join(data)
    if isinstance(data, str):
      match = re.match(r"^#(\s*)(.*)$", data)
      if match is None:
        raise gfapy.FormatError("Comment lines must begin with #\n"+
            "Line: {}".format(data))
      data = ["#", match.group(2), match.group(1)]
    return data

  def _compute_version(self, rt):
    if rt in Construction.RECORD_TYPE_VERSIONS["generic"]:
      self._version = "generic"
    elif rt in Construction.RECORD_TYPE_VERSIONS["different"]:
      if hasattr(self.__class__, "VERSION"):
        self._version = self.__class__.VERSION
      else:
        raise gfapy.RuntimeError(
            "GFA version not specified\n"+
            "Records of type {} ".format(rt)+
            "have different syntax according to the version")
    else:
      for k, v in Construction.RECORD_TYPE_VERSIONS["specific"].items():
        if rt in v:
          self._version = k
          break
    if not self._version:
      self._version = "gfa2"

  def _validate_version(self):
    rt = self.__class__.RECORD_TYPE
    if self._version not in gfapy.VERSIONS:
      raise gfapy.VersionError(
            "GFA specification version unknown ({})".format(self._version))
    else:
      for k, v in Construction.RECORD_TYPE_VERSIONS["specific"].items():
        if rt in v:
          if self._version != k:
            raise gfapy.VersionError(
              "Records of type {} are incompatible ".format(self.record_type)+
              "with version {}".format(self._version))
          return

  @property
  def _n_positional_fields(self):
    return len(self.POSFIELDS)

  def _init_field_value(self, n ,t, s, errmsginfo = None):
    if self.vlevel >= 1:
      s = gfapy.Field._parse_gfa_field(s, t, safe = True, fieldname = n,
                            line = errmsginfo)
    elif t not in self.DELAYED_PARSING_DATATYPES:
      s = gfapy.Field._parse_gfa_field(s, t, safe = (self.vlevel >= 1),
            fieldname = n, line = errmsginfo)
    self._data[n] = s

  def _initialize_positional_fields(self, strings):
    if strings[0] != self.RECORD_TYPE and self.RECORD_TYPE != "\n":
      raise gfapy.FormatError("Record type of records of "+
          "class {} must be {} ({} found)".format(self.__class__,
            self.RECORD_TYPE, strings[0]))
    if self.version is None:
      raise gfapy.AssertionError(
        "Bug found, please report\n"+
        "strings: {}".format(repr(strings)))
    if (self.vlevel >= 1) and (len(strings)-1 < self._n_positional_fields):
      raise gfapy.FormatError(
        "{} positional fields expected, ".format(self._n_positional_fields) +
        "{} found\n{}".format(len(strings)-1, repr(strings)))
    for i, n in enumerate(self.POSFIELDS):
      self._init_field_value(n, self.__class__.DATATYPE[n], strings[i+1],
                       errmsginfo = strings)

  def _initialize_tags(self, strings):
    for i in range(len(self.POSFIELDS)+1, len(strings)):
      self._initialize_tag(*(gfapy.Field._parse_gfa_tag(strings[i])),
          errmsginfo = strings)

  def _initialize_tag(self, n, t, s, errmsginfo = None):
    if (self.vlevel > 0):
      if n in self._data:
        raise gfapy.NotUniqueError(
          "Tag {} found multiple times".format(n))
      elif self._is_predefined_tag(n):
        self._validate_predefined_tag_type(n, t)
      else:
        self._validate_custom_tagname(n)
        self._datatype[n] = t
    else:
      if not self._field_datatype(t):
        self._datatype[n] = t
    self._init_field_value(n, t, s, errmsginfo = errmsginfo)

  @staticmethod
  def _subclass(data, version = None):
    record_type = data[0]
    if record_type and record_type[0] == "#":
      return gfapy.line.Comment
    elif version == "gfa1":
      return gfapy.Line._subclass_GFA1(record_type)
    elif version == "gfa2":
      return gfapy.Line._subclass_GFA2(record_type)
    elif version is None:
      return gfapy.Line._subclass_unknown_version(data)
    else:
      raise gfapy.VersionError(
          "GFA specification version unknown ({})".format(version))

  @staticmethod
  def _subclass_GFA1(record_type):
    if record_type is None:
      raise gfapy.VersionError(
          "gfapy uses virtual records of unknown type for GFA2 only")
    if record_type == "H": return gfapy.line.Header
    elif record_type == "S": return gfapy.line.segment.GFA1
    elif record_type == "#": return gfapy.line.Comment
    elif record_type == "L": return gfapy.line.edge.Link
    elif record_type == "C": return gfapy.line.edge.Containment
    elif record_type == "P": return gfapy.line.group.Path
    else:
      raise gfapy.VersionError(
          "Custom record types are not supported in GFA1: '{}'".format(
            record_type))

  EXTENSIONS = {}
  """Extensions (definition of custom record types) registered by the user."""

  @staticmethod
  def _subclass_GFA2(record_type):
    if record_type == "H": return gfapy.line.Header
    elif record_type == "S": return gfapy.line.segment.GFA2
    elif record_type == "#": return gfapy.line.Comment
    elif record_type == "E": return gfapy.line.edge.GFA2
    elif record_type == "F": return gfapy.line.Fragment
    elif record_type == "G": return gfapy.line.Gap
    elif record_type == "O": return gfapy.line.group.Ordered
    elif record_type == "U": return gfapy.line.group.Unordered
    elif record_type in gfapy.Line.EXTENSIONS:
      return gfapy.Line.EXTENSIONS[record_type]
    else: return gfapy.line.CustomRecord

  @staticmethod
  def _subclass_unknown_version(data):
    record_type = data[0]
    if record_type == "H": return gfapy.line.Header
    elif record_type == "S": return gfapy.line.Segment._subclass(data)
    elif record_type == "#": return gfapy.line.Comment
    elif record_type == "L": return gfapy.line.edge.Link
    elif record_type == "C": return gfapy.line.edge.Containment
    elif record_type == "P": return gfapy.line.group.Path
    elif record_type == "E": return gfapy.line.edge.GFA2
    elif record_type == "F": return gfapy.line.Fragment
    elif record_type == "G": return gfapy.line.Gap
    elif record_type == "O": return gfapy.line.group.Ordered
    elif record_type == "U": return gfapy.line.group.Unordered
    elif record_type in gfapy.Line.EXTENSIONS:
      return gfapy.Line.EXTENSIONS[record_type]
    else: return gfapy.line.CustomRecord

  @classmethod
  def _apply_definitions(cls):
    """
    This avoids calls for fields which are already defined
    """
    cls._define_field_accessors()
    cls._define_field_aliases()
    cls._define_reference_getters()

  @classmethod
  def _define_field_accessors(cls):
    if not cls.PREDEFINED_TAGS:
      cls.PREDEFINED_TAGS = list(set(cls.DATATYPE.keys()) - set(cls.POSFIELDS))
    fieldnames = cls.POSFIELDS + cls.PREDEFINED_TAGS
    if cls.NAME_FIELD and cls.NAME_FIELD not in fieldnames:
      fieldnames.append(cls.NAME_FIELD)
    for fieldname in fieldnames:
      def get_method(self, fieldname):
        return self.get(fieldname)
      def set_method(self, value, fieldname):
        return self._set_existing_field(fieldname, value)
      setattr(cls, fieldname,
          DynamicField(partial(get_method, fieldname = fieldname),
                       partial(set_method, fieldname = fieldname)))
      def try_get_method(self, fieldname):
        return self.try_get(fieldname)
      setattr(cls, "try_get_" + fieldname,
              partialmethod(try_get_method, fieldname = fieldname))

  @classmethod
  def _define_field_aliases(cls):
    if cls.STORAGE_KEY is None and cls.NAME_FIELD is not None:
      cls.STORAGE_KEY = "name"
    if cls.FIELD_ALIAS is None:
      cls.FIELD_ALIAS = {}
    if cls.NAME_FIELD is not None and "name" not in cls.POSFIELDS:
      cls.FIELD_ALIAS["name"] = cls.NAME_FIELD
    for k,v in cls.FIELD_ALIAS.items():
      setattr(cls, k, getattr(cls, v))
      setattr(cls, "try_get_" + k, getattr(cls, "try_get_" + v))

  @classmethod
  def _define_reference_getters(cls):
    for k in cls.DEPENDENT_LINES + cls.OTHER_REFERENCES:
      def get_method(self, k):
        return self._refs.get(k , [])
      def set_method(self, value, k):
        raise gfapy.AttributeError(
            "References collections cannot be set directly")
      setattr(cls, k,
          DynamicField(partial(get_method, k = k),
                       partial(set_method, k = k)))
    def all_references(self):
      return [ item for item in values for values in self._refs ]

  @classmethod
  def register_extension(cls, references=[]):
    # check the definitions
    if isinstance(cls.POSFIELDS,OrderedDict):
      for fieldname, datatype in cls.POSFIELDS.items():
        cls.DATATYPE[fieldname] = datatype
      cls.POSFIELDS = list(cls.POSFIELDS.keys())
    else:
      for posfield in cls.POSFIELDS:
        if posfield not in cls.DATATYPE:
          raise gfapy.RuntimeError("Extension {} ".format(str(cls))+
              "defines no datatype for the positional field {}".format(posfield))
    if hasattr(cls, "TAGS_DATATYPE"):
      for fieldname, datatype in cls.TAGS_DATATYPE.items():
        cls.DATATYPE[fieldname] = datatype
    if not cls.RECORD_TYPE:
      raise gfapy.RuntimeError("Extension {} ".format(str(cls))+
            "does not define the RECORD_TYPE constant")
    if cls.NAME_FIELD is not None:
      gfapy.lines.finders.Finders.RECORDS_WITH_NAME.append(cls.RECORD_TYPE)
    for field, klass, refkey in references:
      if field not in cls.REFERENCE_FIELDS:
        if not cls.REFERENCE_FIELDS:
          cls.REFERENCE_FIELDS = []
        cls.REFERENCE_FIELDS.append(field)
      if refkey not in klass.DEPENDENT_LINES:
        klass.DEPENDENT_LINES.append(refkey)
        klass._define_reference_getters()
      if cls.REFERENCE_INITIALIZERS is None:
        cls.REFERENCE_INITIALIZERS = []
      cls.REFERENCE_INITIALIZERS.append((field, klass, refkey))
    cls._apply_definitions()
    gfapy.Line.EXTENSIONS[cls.RECORD_TYPE] = cls
    gfapy.Line.RECORD_TYPE_VERSIONS["specific"]["gfa2"].append(cls.RECORD_TYPE)




import gfapy
import unittest

class TestLineSegment(unittest.TestCase):

  def test_from_string(self):
    fields = ["S","1","ACGTCACANNN","RC:i:1232","LN:i:11","ab:Z:abcd",
            "FC:i:2321","KC:i:1212"]
    string = "\t".join(fields)
    gfapy.Line(string)
    self.assertIsInstance(gfapy.Line(string), gfapy.line.segment.GFA1)
    self.assertEqual(str(fields[0]), gfapy.Line(string).record_type)
    self.assertEqual(str(fields[1]), gfapy.Line(string).name)
    self.assertEqual(fields[2], gfapy.Line(string).sequence)
    self.assertEqual(1232, gfapy.Line(string).RC)
    self.assertEqual(11, gfapy.Line(string).LN)
    self.assertEqual(2321, gfapy.Line(string).FC)
    self.assertEqual(1212, gfapy.Line(string).KC)
    self.assertEqual("abcd", gfapy.Line(string).ab)
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line(string + "\tH1")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line("S\tH")
    with self.assertRaises(gfapy.FormatError):
      f = fields[:]
      f[2]="!@#?"
      gfapy.Line("\t".join(f), vlevel = 2)
    with self.assertRaises(gfapy.TypeError):
      f=fields[:]
      f[3]="RC:Z:1232"
      gfapy.Line("\t".join(f), version = "gfa1")
    f=["S","2","ACGTCACANNN","LN:i:3"]
    with self.assertRaises(gfapy.InconsistencyError):
      gfapy.Line("\t".join(f), version = "gfa1", vlevel = 2)
    f=["S","2","ACGTCACANNN","LN:i:11"]
    gfapy.Line("\t".join(f))
    f=["S","2","*","LN:i:3"]
    gfapy.Line("\t".join(f))

  def test_forbidden_segment_names(self):
    gfapy.Line("S\tA+B\t*")
    gfapy.Line("S\tA-B\t*")
    gfapy.Line("S\tA,B\t*")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line("S\tA+,B\t*", vlevel = 2)
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line("S\tA-,B\t*", vlevel = 2)

  def test_coverage(self):
    l = gfapy.Line("S\t0\t*\tRC:i:600\tLN:i:100")
    self.assertEqual(6, l.coverage())
    self.assertEqual(6, l.try_get_coverage())
    l = gfapy.Line("S\t0\t*\tRC:i:600")
    self.assertEqual(None, l.coverage())
    self.assertRaises(gfapy.NotFoundError, l.try_get_coverage)
    l = gfapy.Line("S\t0\t*\tLN:i:100")
    self.assertEqual(None, l.coverage())
    self.assertRaises(gfapy.NotFoundError, l.try_get_coverage)
    l = gfapy.Line("S\t0\t*\tFC:i:600\tLN:i:100")
    self.assertEqual(None, l.coverage())
    self.assertRaises(gfapy.NotFoundError, l.try_get_coverage)
    self.assertEqual(6, l.coverage(count_tag = "FC"))
    self.assertEqual(6, l.try_get_coverage(count_tag = "FC"))



class Canonical:

  def is_canonical(self):
    """Checks if a link is expressed in the canonical form.

    Returns:
      bool

    Links can be expressed in two different forms. If a link is
    expressed in the other form, it is converted before storing.

    Note:
      A link is considered canonical in Gfapy if either the from segment name
      is lexicographically smaller than the to segment name, or the two segment
      names are equal, and at least one orientation is positive.

    Note:
      In the special case in which from == to (== s) we have the
      following equivalences:

      ::
          s + s + == s - s -
          s - s - == s + s + (same as previous case)
          s + s - == s + s - (equivalent to itself)
          s - s + == s - s + (equivalent to itself)

      Considering the values on the left, the first one can be taken as
      canonical, the second not, because it can be transformed in the first
      one; the other two values are canonical, as they are only equivalent
      to themselves.
    """
    if self.from_name < self.to_name:
      return True
    elif self.from_name > self.to_name:
      return False
    else:
      return "+" in [self.from_orient, self.to_orient]

  def canonicize(self):
    """The link itself if canonical, the complement link otherwise.

    .. note::
      The method shall be only used before the link is connected to
      a Gfa instance.

    Returns:
      gfapy.line.edge.Link
    """
    if not self.is_canonical():
      return self.complement()



import gfapy
import unittest

class TestAPILinesDestructors(unittest.TestCase):

  def test_delete_links(self):
    gfa = gfapy.Gfa()
    s = ["S\t0\t*", "S\t1\t*", "S\t2\t*"]
    l = "L\t1\t+\t2\t+\t12M"
    c = "C\t1\t+\t0\t+\t12\t12M"
    for line in (s + [l,c]): gfa.append(line)
    self.assertEqual([l], [str(x) for x in gfa.dovetails])
    self.assertEqual([l], [str(x) for x in \
        gfa.segment("1").end_relations("R", ["2", "L"])])
    for x in gfa.segment("1").oriented_relations("+", \
               gfapy.OrientedLine("2", "+")):
      x.disconnect()
    self.assertEqual([], gfa.dovetails)
    self.assertEqual([], gfa.segment("1").end_relations("R", ["2", "L"]))
    self.assertEqual([c], [str(x) for x in gfa.containments])
    self.assertEqual(c,
                 str(gfa.segment("1").relations_to(gfa.segment("0"),
                                    "edges_to_contained")[0]))
    gfa.append(l)
    self.assertNotEqual([], gfa.dovetails)
    for x in gfa.segment("1").oriented_relations("+", \
               gfapy.OrientedLine("2", "+")):
      x.disconnect()
    self.assertEqual([], gfa.dovetails)

  def test_delete_containments(self):
    gfa = gfapy.Gfa()
    s = ["S\t0\t*", "S\t1\t*", "S\t2\t*"]
    l = "L\t1\t+\t2\t+\t12M"
    c = "C\t1\t+\t0\t+\t12\t12M"
    for line in (s + [l,c]): gfa.append(line)
    for x in gfa.segment("1").relations_to(gfa.segment("0"), "edges_to_contained"):
      x.disconnect()
    self.assertEqual([], gfa.containments)
    self.assertEqual(0, len(gfa.segment("1").relations_to("0",
                                                     "edges_to_contained")))
    gfa.append(c)
    self.assertNotEqual([], gfa.containments)
    self.assertEqual(c, str(gfa.segment("1").relations_to("0",
                                                   "edges_to_contained")[0]))
    for x in gfa.segment("1").relations_to(gfa.segment("0"), "edges_to_contained"):
      x.disconnect()
    self.assertEqual([], gfa.containments)

  def test_delete_segment(self):
    gfa = gfapy.Gfa()
    gfa.append("H\tVN:Z:1.0")
    s = ["S\t0\t*", "S\t1\t*", "S\t2\t*"]
    l = "L\t1\t+\t2\t+\t12M"
    c = "C\t1\t+\t0\t+\t12\t12M"
    p = "P\t4\t2+,0-\t12M"
    for line in (s + [l,c,p]): gfa.append(line)
    self.assertEqual(set(s), set([str(x) for x in gfa.segments]))
    self.assertEqual(set(["0", "1", "2"]), set(gfa.segment_names))
    self.assertEqual([l], [str(x) for x in gfa.dovetails if not x.virtual])
    self.assertEqual([c], [str(x) for x in gfa.containments])
    self.assertEqual([p], [str(x) for x in gfa.paths])
    self.assertEqual(["4"], gfa.path_names)
    gfa.segment("0").disconnect()
    self.assertEqual(set([s[1],s[2]]), set([str(x) for x in gfa.segments]))
    self.assertEqual(set(["1", "2"]), set(gfa.segment_names))
    self.assertEqual([l], [str(x) for x in gfa.dovetails if not x.virtual])
    self.assertEqual([], [str(x) for x in gfa.containments])
    self.assertEqual([], [str(x) for x in gfa.paths])
    self.assertEqual([], gfa.path_names)
    gfa.segment("1").disconnect()
    self.assertEqual([s[2]], [str(x) for x in gfa.segments])
    self.assertEqual([], gfa.dovetails)
    gfa.rm("2")
    self.assertEqual([], gfa.segments)




import gfapy
import unittest

class TestAPIReferencesEdgesGFA1(unittest.TestCase):

  def test_links_references(self):
    g = gfapy.Gfa()
    lab = gfapy.Line("L\ta\t+\tb\t+\t*")
    self.assertEqual("a", lab.from_segment)
    self.assertEqual("b", lab.to_segment)
    g.append(lab)
    sa = gfapy.Line("S\ta\t*")
    g.append(sa)
    sb = gfapy.Line("S\tb\t*")
    g.append(sb)
    self.assertEqual(sa, lab.from_segment)
    self.assertEqual(sb, lab.to_segment)
    lab.disconnect()
    self.assertEqual("a", lab.from_segment)
    self.assertEqual("b", lab.to_segment)
    # disconnection of segment cascades on links
    g.append(lab)
    assert(lab.is_connected())
    self.assertEqual(sa, lab.from_segment)
    sa.disconnect()
    assert(not lab.is_connected())
    self.assertEqual("a", lab.from_segment)

  def test_links_backreferences(self):
    g = gfapy.Gfa()
    sa = gfapy.Line("S\ta\t*")
    g.append(sa)
    # links
    s = {}; l = {}
    for name in ["b", "c", "d", "e", "f", "g", "h", "i"]:
      s[name] = gfapy.Line("S\t{}\t*".format(name))
      g.append(s[name])
    for name in \
         ["a+b+", "a+c-", "a-d+", "a-e-", "f+a+", "g+a-", "h-a+", "i-a-"]:
      l[name] = gfapy.Line("\t".join(list("L"+name+"*")))
      g.append(l[name])
    # dovetails_[LR]()
    self.assertEqual([l["a+b+"], l["a+c-"],
                  l["g+a-"], l["i-a-"]], sa.dovetails_R)
    self.assertEqual([l["a-d+"], l["a-e-"],
                  l["f+a+"], l["h-a+"]], sa.dovetails_L)
    # dovetails()
    self.assertEqual(sa.dovetails_R, sa.dovetails_of_end("R"))
    self.assertEqual(sa.dovetails_L, sa.dovetails_of_end("L"))
    self.assertEqual(sa.dovetails_L + sa.dovetails_R, sa.dovetails)
    # neighbours
    self.assertEqual(set(["b", "c", "d", "e", "f", "g", "h", "i"]),
                     set([x.name for x in sa.neighbours]))
    # gfa2 specific collections are empty in gfa1
    self.assertEqual([], sa.gaps)
    self.assertEqual([], sa.fragments)
    self.assertEqual([], sa.internals)
    # upon disconnection
    sa.disconnect()
    self.assertEqual([], sa.dovetails_R)
    self.assertEqual([], sa.dovetails_R)
    self.assertEqual([], sa.dovetails_of_end("L"))
    self.assertEqual([], sa.dovetails_of_end("R"))
    self.assertEqual([], sa.dovetails)
    self.assertEqual([], sa.neighbours)

  def test_containments_references(self):
    g = gfapy.Gfa()
    cab = gfapy.Line("C\ta\t+\tb\t+\t10\t*")
    self.assertEqual("a", cab.from_segment)
    self.assertEqual("b", cab.to_segment)
    sa = gfapy.Line("S\ta\t*")
    g.append(sa)
    sb = gfapy.Line("S\tb\t*")
    g.append(sb)
    g.append(cab)
    self.assertEqual(sa, cab.from_segment)
    self.assertEqual(sb, cab.to_segment)
    cab.disconnect()
    self.assertEqual("a", cab.from_segment)
    self.assertEqual("b", cab.to_segment)
    # disconnection of segment cascades on containments
    g.append(cab)
    assert(cab.is_connected())
    self.assertEqual(sa, cab.from_segment)
    sa.disconnect()
    assert(not cab.is_connected())
    self.assertEqual("a", cab.from_segment)

  def test_containments_backreferences(self):
    g = gfapy.Gfa()
    sa = gfapy.Line("S\ta\t*")
    g.append(sa)
    # containments:
    s = {}; c = {}
    for name in ["b", "c", "d", "e", "f", "g", "h", "i"]:
      s[name] = gfapy.Line("S\t"+"{}".format(name)+"\t*")
      g.append(s[name])
    for name in \
        ["a+b+", "a+c-", "a-d+", "a-e-", "f+a+", "g+a-", "h-a+", "i-a-"]:
      c[name] = gfapy.Line("\t".join(list("C{}9*".format(name))))
      g.append(c[name])
    # edges to contained/containers
    self.assertEqual([c["a+b+"], c["a+c-"], c["a-d+"], c["a-e-"]],
                     sa.edges_to_contained)
    self.assertEqual([c["f+a+"], c["g+a-"], c["h-a+"], c["i-a-"]],
                     sa.edges_to_containers)
    # containments
    self.assertEqual(sa.edges_to_contained + sa.edges_to_containers,
                     sa.containments)
    # contained/containers
    self.assertEqual([s["b"], s["c"], s["d"], s["e"]], sa.contained)
    self.assertEqual([s["f"], s["g"], s["h"], s["i"]], sa.containers)
    # gfa2 specific collections are empty in gfa1
    self.assertEqual([], sa.gaps)
    self.assertEqual([], sa.fragments)
    self.assertEqual([], sa.internals)
    # upon disconnection
    sa.disconnect()
    self.assertEqual([], sa.edges_to_contained)
    self.assertEqual([], sa.edges_to_containers)
    self.assertEqual([], sa.containments)
    self.assertEqual([], sa.contained)
    self.assertEqual([], sa.containers)




from ..gfa2.references import References
from ..gfa2.same_id import SameID
from ..unordered.references import References as UnorderedReferences
from ..unordered.induced_set import InducedSet
from ..group import Group

class Unordered(UnorderedReferences, InducedSet, References, SameID, Group):
  """An unordered group line of a GFA2 file"""
  RECORD_TYPE = "U"
  POSFIELDS = ["pid", "items"]
  FIELD_ALIAS = {"name" : "pid"}
  DATATYPE = {
    "pid" : "optional_identifier_gfa2",
    "items" : "identifier_list_gfa2",
  }
  NAME_FIELD = "pid"
  REFERENCE_FIELDS = ["items"]
  DEPENDENT_LINES = ["sets"]

Unordered._apply_definitions()



import gfapy
import unittest

class TestAPIPositionals(unittest.TestCase):

  s = {
    "S1": "S\t1\t*",
    "L": "L\t1\t+\t2\t+\t*",
    "C": "C\t1\t+\t2\t+\t10\t*",
    "P": "P\tx\t1+,2+\t*",
    "S2": "S\t2\t100\t*",
    "E": "E\t*\t1+\t2+\t10\t20\t30\t40\t*",
    "F": "F\t1\t5+\t11\t21\t31\t41\t*",
    "G": "G\t*\t1+\t2+\t1000\t1",
    "U": "U\t*\t1 2 3",
    "O": "O\t*\t1+ 2+ 3+",
  }
  f = {k:v.split("\t") for k, v in s.items()}
  l = {k:gfapy.Line(v) for k, v in s.items()}

  fieldnames = {
    "S1":["name", "sequence"],
    "L" :["from_segment", "from_orient", "to_segment", "to_orient", "overlap"],
    "C" :["from_segment", "from_orient", "to_segment", "to_orient", "pos", "overlap"],
    "P" :["path_name", "segment_names", "overlaps"],
    "S2":["sid", "slen", "sequence"],
    "E" :["eid", "sid1", "sid2", "beg1", "end1", "beg2", "end2", "alignment"],
    "F" :["sid", "external", "s_beg", "s_end", "f_beg", "f_end", "alignment"],
    "G" :["gid", "sid1", "sid2", "disp", "var"],
    "U" :["pid", "items"],
    "O" :["pid", "items"],
  }

  # alternative values to set tests
  v1 = {
    "S1":{"name":"sx", "sequence":"accg"},
    "L":{"from_segment":"a1", "from_orient":"-", "to_segment":"a2", "to_orient":"-",
           "overlap": gfapy.Alignment("12M")},
    "C":{"from_segment":"cx", "from_orient":"-", "to_segment":"cy", "to_orient":"-",
           "pos":123, "overlap": gfapy.Alignment("120M")},
    "P":{"path_name":"px", "segment_names":[gfapy.OrientedLine("x","+"), gfapy.OrientedLine("y","-")],
           "overlaps":[gfapy.Alignment("10M")]},
    "S2":{"sid":"s2s", "slen":999, "sequence":"gggg"},
    "E" :{"eid":"e2e", "sid1":gfapy.OrientedLine("s2s","-"),
            "sid2":gfapy.OrientedLine("t2t","-"),
            "beg1":0, "end1":gfapy.LastPos("100$"),
            "beg2":10, "end2":gfapy.LastPos("110$"),
            "alignment":gfapy.Alignment("10M1I10M1D80M")},
    "F" :{"sid":"s2s", "external":gfapy.OrientedLine("ex2ex","-"),
            "s_beg":0, "s_end":gfapy.LastPos("100$"),
            "f_beg":10, "f_end":gfapy.LastPos("110$"),
            "alignment":gfapy.Alignment("10M1I10M1D80M")},
    "G" :{"gid":"g2g", "sid1":gfapy.OrientedLine("s2s","+"), "sid2":gfapy.OrientedLine("t2t","-"),
            "disp":2000, "var":100},
    "O" :{"pid":"O100", "items":[gfapy.OrientedLine("x1","+"),
                                      gfapy.OrientedLine("x2","+"),
                                      gfapy.OrientedLine("x3","-")]},
    "U" :{"pid":"U100", "items":["x1", "x2", "x3"]},
  }
  v2 = {
    "S1":{"name":"xs", "sequence":"aggc"},
    "L":{"from_segment":"a5", "from_orient":"+", "to_segment":"a7", "to_orient":"+",
           "overlap":gfapy.Alignment("9M3I3M")},
    "C":{"from_segment":"cp", "from_orient":"+", "to_segment":"cl", "to_orient":"+",
           "pos":213, "overlap":gfapy.Alignment("110M4D10M")},
    "P":{"path_name":"pu", "segment_names":[gfapy.OrientedLine("k","-"),
           gfapy.OrientedLine("l","+")], "overlaps":[gfapy.Alignment("11M")]},
    "S2":{"sid":"s4s", "slen":1999, "sequence":"aaaa"},
    "E" :{"eid":"e4e", "sid1":gfapy.OrientedLine("s4s","+"),
            "sid2":gfapy.OrientedLine("t4t","+"),
            "beg1":10, "end1":gfapy.LastPos("110$"),
            "beg2":0, "end2":gfapy.LastPos("100$"),
            "alignment":gfapy.Alignment("10M1I20M1D80M")},
    "F" :{"sid":"s4s", "external":gfapy.OrientedLine("ex4ex", "+"),
            "s_beg":10, "s_end":gfapy.LastPos("110$"),
            "f_beg":0, "f_end":gfapy.LastPos("100$"),
            "alignment":gfapy.Alignment("10M1I20M1D80M")},
    "G" :{"gid":"g4g", "sid1":gfapy.OrientedLine("s4s","-"), "sid2":gfapy.OrientedLine("t4t","+"),
            "disp":3000, "var":200},
    "O" :{"pid":"O200", "items":[gfapy.OrientedLine("x7","-"),
                                      gfapy.OrientedLine("x6","+"),
                                      gfapy.OrientedLine("x3","+")]},
    "U" :{"pid":"U200", "items":["x6", "x7", "x4"]},
  }
  aliases = {
      "S1":{"name":"sid"}, "P":{"path_name":"name"},
      "S2":{"sid":"name"}, "E":{"eid":"name"}, "G":{"gid":"name"},
      "U":{"pid":"name"}, "O":{"pid":"name"},
      "L":{"from_segment": "from", "to_segment": "to"},
      "C":{"from_segment":"container", "from_orient":"container_orient",
             "to_segment":"contained", "to_orient":"contained_orient"}
  }

  def test_number_of_positionals(self):
    for rt, fields in TestAPIPositionals.f.items():
      gfapy.Line(fields) # nothing raised
      too_less = fields.copy(); too_less.pop()
      with self.assertRaises(gfapy.FormatError): gfapy.Line(too_less)
      too_many = fields.copy(); too_many.append("*")
      with self.assertRaises(gfapy.FormatError): gfapy.Line(too_many)

  def test_positional_fieldnames(self):
    for rt, line in TestAPIPositionals.l.items():
      self.assertEqual(TestAPIPositionals.fieldnames[rt], line.positional_fieldnames)

  def test_field_getters_and_setters(self):
    for rt, fn_list in TestAPIPositionals.fieldnames.items():
      for i, fn in enumerate(fn_list):
        i+=1 # skip record_type
        # field_to_s()
        self.assertEqual(TestAPIPositionals.f[rt][i], TestAPIPositionals.l[rt].field_to_s(fn))
        # validate_field/validate
        TestAPIPositionals.l[rt].validate_field(fn) # nothing raised
        TestAPIPositionals.l[rt].validate # nothing raised
        # fieldname() == get(fieldname)
        self.assertEqual(getattr(TestAPIPositionals.l[rt], fn), TestAPIPositionals.l[rt].get(fn))
        # fieldname=() and fieldname()
        l = TestAPIPositionals.l[rt].clone()
        setattr(l,fn,TestAPIPositionals.v1[rt][fn])
        self.assertEqual(TestAPIPositionals.v1[rt][fn], getattr(l, fn))
        # set() and get()
        l.set(fn, TestAPIPositionals.v2[rt][fn])
        self.assertEqual(TestAPIPositionals.v2[rt][fn], l.get(fn))

  def test_aliases(self):
    for rt, aliasmap in TestAPIPositionals.aliases.items():
      for orig, al in aliasmap.items():
        # get(orig) == get(alias)
        self.assertEqual(getattr(TestAPIPositionals.l[rt], orig), getattr(TestAPIPositionals.l[rt],al))
        self.assertEqual(TestAPIPositionals.l[rt].get(orig), TestAPIPositionals.l[rt].get(al))
        # validate_field/validate
        TestAPIPositionals.l[rt].validate_field(al) # nothing raised
        TestAPIPositionals.l[rt].validate # nothing raised
        # field_to_s(orig) == field_to_s(alias)
        self.assertEqual(TestAPIPositionals.l[rt].field_to_s(orig), TestAPIPositionals.l[rt].field_to_s(al))
        # set(al, value) + get(orig)
        l = TestAPIPositionals.l[rt].clone()
        self.assertNotEqual(TestAPIPositionals.v1[rt][orig], getattr(l,orig))
        l.set(al, TestAPIPositionals.v1[rt][orig])
        self.assertEqual(TestAPIPositionals.v1[rt][orig], getattr(l,orig))
        # alias=value + orig()
        self.assertNotEqual(TestAPIPositionals.v2[rt][orig], getattr(l,orig))
        setattr(l, al, TestAPIPositionals.v2[rt][orig])
        self.assertEqual(TestAPIPositionals.v2[rt][orig], getattr(l,orig))
        # set(orig, value) + get(alias)
        self.assertNotEqual(TestAPIPositionals.v1[rt][orig], getattr(l,al))
        l.set(orig, TestAPIPositionals.v1[rt][orig])
        self.assertEqual(TestAPIPositionals.v1[rt][orig], getattr(l,al))
        # orig=value + alias()
        self.assertNotEqual(TestAPIPositionals.v2[rt][orig], getattr(l,al))
        setattr(l, orig, TestAPIPositionals.v2[rt][orig])
        self.assertEqual(TestAPIPositionals.v2[rt][orig], getattr(l,al))

  def test_array_fields(self):
    assert(isinstance(TestAPIPositionals.l["P"].segment_names, list))
    assert(isinstance(TestAPIPositionals.l["P"].segment_names[0], gfapy.OrientedLine))
    assert(isinstance(TestAPIPositionals.l["P"].overlaps, list))
    assert(isinstance(TestAPIPositionals.l["P"].overlaps[0], gfapy.AlignmentPlaceholder))
    assert(isinstance(TestAPIPositionals.l["O"].items, list))
    assert(isinstance(TestAPIPositionals.l["O"].items[0], gfapy.OrientedLine))
    assert(isinstance(TestAPIPositionals.l["U"].items, list))
    assert(isinstance(TestAPIPositionals.l["U"].items[0], str))

  def test_orientation(self):
    # orientation is symbol
    self.assertEqual("+", TestAPIPositionals.l["L"].from_orient)
    self.assertEqual("+", TestAPIPositionals.l["L"].to_orient)
    # invert
    self.assertEqual("-", gfapy.invert(TestAPIPositionals.l["L"].to_orient))
    self.assertEqual("+", gfapy.invert("-"))
    self.assertEqual("-", gfapy.invert("+"))
    # string representation
    self.assertEqual("+", TestAPIPositionals.l["L"].field_to_s("from_orient"))
    # assigning the string representation
    l = TestAPIPositionals.l["L"].clone()
    l.from_orient = "+"
    self.assertEqual("+", l.from_orient)
    self.assertEqual("-", gfapy.invert(l.from_orient))
    # non "+"/"-" symbols is an error
    with self.assertRaises(gfapy.FormatError):
      l.from_orient = "x"
      l.validate()
    # only "+"/"-" and their string representations are accepted
    with self.assertRaises(gfapy.FormatError):
      l.from_orient = "x"
      l.validate()
    with self.assertRaises(gfapy.FormatError):
      l.from_orient = 1
      l.validate()

  def test_oriented_segment(self):
    os = TestAPIPositionals.l["P"].segment_names[0]
    # getter methods
    self.assertEqual("1", os.line)
    self.assertEqual("+", os.orient)
    # invert
    self.assertEqual("1", os.inverted().line)
    self.assertEqual("-", os.inverted().orient)
    self.assertEqual("-", gfapy.invert(os.orient))
    # setter methods
    os.line = "one"
    os.orient = "-"
    self.assertEqual("one", os.line)
    self.assertEqual("-", os.orient)
    # name
    self.assertEqual("one", os.name)
    os.line = TestAPIPositionals.l["S1"]
    self.assertEqual(TestAPIPositionals.l["S1"], os.line)
    self.assertEqual(TestAPIPositionals.l["S1"].name, os.name)

  def test_sequence(self):
    # placeholder
    assert(gfapy.is_placeholder(TestAPIPositionals.l["S1"].sequence))
    assert(gfapy.is_placeholder(TestAPIPositionals.l["S2"].sequence))
    s = TestAPIPositionals.l["S1"].clone()
    s.sequence = "ACCT"
    assert(not gfapy.is_placeholder(s.sequence))
    # sequence is string
    self.assertEqual("ACCT", s.sequence)
    # rc
    self.assertEqual("AGGT", gfapy.sequence.rc(s.sequence))
    # GFA2 allowed alphabet is larger than GFA1
    s.validate # nothing raised
    s.sequence = ";;;{}"
    with self.assertRaises(gfapy.FormatError): s.validate()
    s = TestAPIPositionals.l["S2"].clone()
    s.sequence = ";;;{}"
    s.validate() # nothing raised
    # Sequence
    assert(isinstance(gfapy.sequence.Sequence("*"),gfapy.Placeholder))
    assert(isinstance(gfapy.sequence.Sequence("ACG"),str))

  def test_sequence_rc(self):
    self.assertEqual("gcatcgatcgt",gfapy.sequence.rc("acgatcgatgc"))
    # case
    self.assertEqual("gCaTCgatcgt",gfapy.sequence.rc("acgatcGAtGc"))
    # wildcards
    self.assertEqual("gcatcnatcgt",gfapy.sequence.rc("acgatngatgc"))
    self.assertEqual("gcatcYatcgt",gfapy.sequence.rc("acgatRgatgc"))
    # RNA
    self.assertEqual("gcaucgaucgu",gfapy.sequence.rc("acgaucgaugc",rna=True))
    self.assertEqual("===.",gfapy.sequence.rc(".==="))
    # valid
    with self.assertRaises(gfapy.ValueError): gfapy.sequence.rc("acgatXgatgc")
    gfapy.sequence.rc("acgatXgatgc",valid=True) # nothing raised
    # placeholder
    self.assertEqual("*",gfapy.sequence.rc("*"))
    with self.assertRaises(gfapy.ValueError): gfapy.sequence.rc("**")




import gfapy
import re

def decode(string):
  validate_encoded(string)
  return string

unsafe_decode = decode

def validate_encoded(string):
  if not re.match(r"^[!-~]$", string):
    raise gfapy.FormatError(
        "{} is not a single printable character string".format(repr(string)))

def validate_decoded(string):
  return validate_encoded(string)

def unsafe_encode(obj):
  return str(obj)

def encode(obj):
  if not isinstance(obj, str):
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str)")
  validate_encoded(obj)
  return obj



import gfapy
import json
import re

def unsafe_decode(string):
  return json.loads(string)

def decode(string):
  validate_all_printable(string)
  return unsafe_decode(string)

def validate_encoded(string):
  # both regex and JSON parse are necessary,
  # because string can be invalid JSON and
  # JSON can contain forbidden chars (non-printable)
  validate_all_printable(string)
  try:
    json.loads(string)
  except Exception as err:
    raise Exception(
    "{} is not a valid JSON string\n".format(repr(string))+
    "json.loads raised a {} exception\n".format(err.__class__.__name__)+
    "error message: {}").format(str(err)) from err

def validate_decoded(obj):
  if isinstance(obj, gfapy.FieldArray):
    obj.validate()
  elif isinstance(obj, list) or isinstance(obj, dict):
    string = encode(obj)
    validate_all_printable(string)
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__) +
      "(accepted classes: list, dict, gfapy.FieldArray)")

def unsafe_encode(obj):
  return json.dumps(obj)

def encode(obj):
  if isinstance(obj, str):
    validate_encoded(obj)
    return obj
  elif isinstance(obj, list) or isinstance(obj, dict):
    string = json.dumps(obj)
    validate_all_printable(string)
    return string
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__) +
      "(accepted classes: list, dict, gfapy.FieldArray)")

def validate_all_printable(string):
  if not re.match("^[ !-~]+$", string):
    raise gfapy.FormatError(
      "{} is not a valid JSON field\n".format(repr(string))+
      "(it contains newlines, tabs and/or non-printable characters)")



from ..common.from_to import FromTo
from ..common.alignment_type import AlignmentType
from ..gfa2.to_gfa1 import ToGFA1
from ..gfa2.alignment_type import AlignmentType as GFA2_AlignmentType
from ..gfa2.references import References
from ..gfa2.other import Other
from ..gfa2.validation import Validation
from ..edge import Edge

class GFA2(Other, References, GFA2_AlignmentType, AlignmentType, FromTo,
           ToGFA1, Validation, Edge):
  """An edge line of a GFA2 file."""

  RECORD_TYPE = "E"
  POSFIELDS = ["eid", "sid1", "sid2", "beg1", "end1", "beg2", "end2",
               "alignment"]
  PREDEFINED_TAGS = ["TS"]
  DATATYPE = {
    "eid" : "optional_identifier_gfa2",
    "sid1" : "oriented_identifier_gfa2",
    "sid2" : "oriented_identifier_gfa2",
    "beg1" : "position_gfa2",
    "end1" : "position_gfa2",
    "beg2" : "position_gfa2",
    "end2" : "position_gfa2",
    "alignment" : "alignment_gfa2",
    "TS" : "i",
  }
  NAME_FIELD = "eid"
  FIELD_ALIAS = { "name" : "eid" }
  REFERENCE_FIELDS = ["sid1", "sid2"]
  BACKREFERENCE_RELATED_FIELDS = ["beg1", "end1", "beg2", "end2"]
  DEPENDENT_LINES = ["paths", "sets"]

GFA2._apply_definitions()




import unittest
import gfapy

class TestUnitGfaLines(unittest.TestCase):

  def test_register_line_merge(self):
    g = gfapy.Gfa(version="gfa1")
    l = gfapy.line.Header({"xx": 1}, version="gfa1")
    l._gfa = g
    g._register_line(l)
    self.assertEqual(1, g.header.xx)
    self.assertRaises(gfapy.AssertionError, g._unregister_line, l)

  def test_register_line_name_present(self):
    g = gfapy.Gfa(version="gfa1")
    l = gfapy.line.segment.GFA1({"name": "sx"}, version="gfa1")
    l._gfa = g
    g._register_line(l)
    self.assertEqual([l], g.segments)
    self.assertEqual(l, g.line("sx"))
    self.assertEqual(["sx"], g.segment_names)
    g._unregister_line(l)
    self.assertEqual([], g.segments)
    self.assertEqual(None, g.line("sx"))
    self.assertEqual([], g.segment_names)

  def test_register_line_name_absent(self):
    g = gfapy.Gfa(version="gfa2")
    l = gfapy.line.edge.GFA2({"eid": gfapy.Placeholder()},
                             version="gfa2")
    l._gfa = g
    g._register_line(l)
    self.assertEqual([l], g.edges)
    self.assertEqual([], g.edge_names)
    g._unregister_line(l)
    self.assertEqual([], g.edges)

  def test_register_line_external(self):
    g = gfapy.Gfa(version="gfa2")
    l = gfapy.line.Fragment({"external": gfapy.OrientedLine("x","+")},
                                  version="gfa2")
    l._gfa = g
    g._register_line(l)
    self.assertEqual([l], g.fragments)
    self.assertEqual([l], g.fragments_for_external("x"))
    self.assertEqual(["x"], g.external_names)
    g._unregister_line(l)
    self.assertEqual([], g.fragments)
    self.assertEqual([], g.fragments_for_external("x"))
    self.assertEqual([], g.external_names)

  def test_register_line_unnamed(self):
    g = gfapy.Gfa(version="gfa1")
    l = gfapy.line.edge.Link({}, version="gfa1")
    l._gfa = g
    g._register_line(l)
    self.assertEqual([l], g.dovetails)
    g._unregister_line(l)
    self.assertEqual([], g.dovetails)


